#ifndef MMTRACE
#include <stdio.h>
#include <string.h>
#endif
#include "shader_common.h"
#include "nv20_shader.h"

unsigned int NVVP_TX_VOP_COUNT = 16;
unsigned int NVVP_TX_SOP_COUNT = 16;
struct _op_xlat NVVP_TX_VOP[32];
struct _op_xlat NVVP_TX_SOP[32];

ShaderSwzComp NV20VP_TX_SWIZZLE[4] = { SSC_X, SSC_Y, SSC_Z, SSC_W };

ShaderRegSwz
NV20VPTXSwizzle(int hwswz)
{
	ShaderRegSwz swz;

	swz.x = NV20VP_TX_SWIZZLE[(hwswz & 0xC0) >> 6];
	swz.y = NV20VP_TX_SWIZZLE[(hwswz & 0x30) >> 4];
	swz.z = NV20VP_TX_SWIZZLE[(hwswz & 0x0C) >> 2];
	swz.w = NV20VP_TX_SWIZZLE[(hwswz & 0x03) >> 0];

	return swz;
}

static int
NV20VPHasMergedInst(ShaderFunc *shader)
{
	if (shader->GetOpcodeHW(shader, 0) != NV20_VP_INST_OPCODE_NOP &&
		shader->GetOpcodeHW(shader, 1) != NV20_VP_INST_OPCODE_NOP)
			printf("\n\n# *****both opcode fields have values - PLEASE REPORT*****\n");
	return 0;
}

static int
NV20VPIsLastInst(ShaderFunc *shader)
{
	return ((shader->inst[3] & (1<<0)) ? 1 : 0);
}

static int
NV20VPGetOffsetNext(ShaderFunc *shader)
{
	return 4;
}

static struct _op_xlat *
NV20VPGetOPTXRec(ShaderFunc *shader, int merged)
{
	struct _op_xlat *opr;
	int op;

	if (shader->GetOpcodeSlot(shader, merged)) {
		opr = NVVP_TX_SOP;
		op = shader->GetOpcodeHW(shader, 1);
		if (op >= NVVP_TX_SOP_COUNT)
			return NULL;
	} else {
		opr = NVVP_TX_VOP;
		op = shader->GetOpcodeHW(shader, 0);
		if (op >= NVVP_TX_VOP_COUNT)
			return NULL;
	}

	if (opr[op].SOP == SOP_UNKNOWN)
		return NULL;
	return &opr[op];
}

static int
NV20VPGetOpcodeSlot(ShaderFunc *shader, int merged)
{
	if (shader->HasMergedInst(shader))
		return merged;
	if (shader->GetOpcodeHW(shader, 0) == NV20_VP_INST_OPCODE_NOP)
		return 1;
	return 0;
}

static ShaderOpcode
NV20VPGetOpcode(ShaderFunc *shader, int merged)
{
	struct _op_xlat *opr;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr)
		return SOP_UNKNOWN;

	return opr->SOP;
}

static ShaderOpcode
NV20VPGetOpcodeHW(ShaderFunc *shader, int slot)
{
	if (slot)
		return (shader->inst[1] & NV20_VP_INST_SCA_OPCODE_MASK)
			>> NV20_VP_INST_SCA_OPCODE_SHIFT;
	return (shader->inst[1] & NV20_VP_INST_VEC_OPCODE_MASK)
		>> NV20_VP_INST_VEC_OPCODE_SHIFT;
}

static ShaderRegFile
NV20VPGetDestFile(ShaderFunc *shader, int merged)
{
	switch (shader->GetOpcode(shader, merged)) {
	case SOP_ARL:
		return SRF_ADDRESS;
	default:
	/*FIXME: This probably isn't correct.. */
		if ((shader->inst[3] & NV20_VP_INST_DEST_WRITEMASK_MASK) == 0)
			return SRF_TEMP;
		return SRF_RESULT;
	}
}

static unsigned int
NV20VPGetDestID(ShaderFunc *shader, int merged)
{
	int id;

	switch (shader->GetDestFile(shader, merged)) {
	case SRF_RESULT:
		id = ((shader->inst[3] & NV20_VP_INST_DEST_MASK)
				>> NV20_VP_INST_DEST_SHIFT);
		switch (id) {
		case NV20_VP_INST_DEST_POS  : return SR_POSITION;
		case NV20_VP_INST_DEST_COL0 : return SR_COL0;
		case NV20_VP_INST_DEST_COL1 : return SR_COL1;
		case NV20_VP_INST_DEST_TC(0): return SR_TEXCOORD0;
		case NV20_VP_INST_DEST_TC(1): return SR_TEXCOORD1;
		case NV20_VP_INST_DEST_TC(2): return SR_TEXCOORD2;
		case NV20_VP_INST_DEST_TC(3): return SR_TEXCOORD3;
		default:
			return -1;
		}
	case SRF_ADDRESS:
		return 0;
	case SRF_TEMP:
		id = ((shader->inst[3] & NV20_VP_INST_DEST_TEMP_ID_MASK)
				>> NV20_VP_INST_DEST_TEMP_ID_SHIFT);
		return id;
	default:
		return -1;
	}
}

static unsigned int
NV20VPGetDestMask(ShaderFunc *shader, int merged)
{
	int hwmask, mask = 0;

	/* Special handling for ARL - hardware only supports a
	 * 1-component address reg
	 */
	if (shader->GetOpcode(shader, merged) == SOP_ARL)
		return SMASK_X;

	if (shader->GetDestFile(shader, merged) == SRF_RESULT)
		hwmask = (shader->inst[3] & NV20_VP_INST_DEST_WRITEMASK_MASK)
			>> NV20_VP_INST_DEST_WRITEMASK_SHIFT;
	else
		if (shader->GetOpcodeSlot(shader, merged))
			hwmask = (shader->inst[3] & NV20_VP_INST_STEMP_WRITEMASK_MASK)
				>> NV20_VP_INST_STEMP_WRITEMASK_SHIFT;
		else
			hwmask = (shader->inst[3] & NV20_VP_INST_VTEMP_WRITEMASK_MASK)
				>> NV20_VP_INST_VTEMP_WRITEMASK_SHIFT;

	if (hwmask & (1<<3)) mask |= SMASK_X;
	if (hwmask & (1<<2)) mask |= SMASK_Y;
	if (hwmask & (1<<1)) mask |= SMASK_Z;
	if (hwmask & (1<<0)) mask |= SMASK_W;

	return mask;
}

static unsigned int
NV20VPGetSourceHW(ShaderFunc *shader, int merged, int pos)
{
	struct _op_xlat *opr;
	unsigned int src;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr)
		return -1;

	switch (opr->srcpos[pos]) {
	case 0:
		src = ((shader->inst[1] & NV20_VP_INST_SRC0H_MASK)
				>> NV20_VP_INST_SRC0H_SHIFT)
			<< NV20_VP_SRC0_HIGH_SHIFT;
		src |= ((shader->inst[2] & NV20_VP_INST_SRC0L_MASK)
				>> NV20_VP_INST_SRC0L_SHIFT);
		break;
	case 1:
		src = ((shader->inst[2] & NV20_VP_INST_SRC1_MASK)
				>> NV20_VP_INST_SRC1_SHIFT);
		break;
	case 2:
		src = ((shader->inst[2] & NV20_VP_INST_SRC2H_MASK)
				>> NV20_VP_INST_SRC2H_SHIFT)
			<< NV20_VP_SRC2_HIGH_SHIFT;
		src |= ((shader->inst[3] & NV20_VP_INST_SRC2L_MASK)
				>> NV20_VP_INST_SRC2L_SHIFT);
		break;
	default:
		src = -1;
	}

	return src;
}

static ShaderRegFile
NV20VPGetSourceFile(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	struct _op_xlat *opr;
	int file;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1)
		return -1;

	switch (opr->srcpos[pos]) {
	case SPOS_ADDRESS: return SRF_ADDRESS;
	default:
		src  = NV20VPGetSourceHW(shader, merged, pos);
		file = (src & NV20_VP_SRC_REG_TYPE_MASK) >> NV20_VP_SRC_REG_TYPE_SHIFT;

		switch (file) {
		case NV20_VP_SRC_REG_TYPE_TEMP : return SRF_TEMP;
		case NV20_VP_SRC_REG_TYPE_INPUT: return SRF_ATTRIB;
		case NV20_VP_SRC_REG_TYPE_CONST: return SRF_CONST;
		default:
			return SRF_UNKNOWN;
		}
	}
}

static int
NV20VPGetSourceID(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;

	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_TEMP:
		src = shader->GetSourceHW(shader, merged, pos);
		return ((src & NV20_VP_SRC_REG_TEMP_ID_MASK) >>
				NV20_VP_SRC_REG_TEMP_ID_SHIFT);
	case SRF_CONST:
		return ((shader->inst[1] & NV20_VP_INST_CONST_SRC_MASK)
				>> NV20_VP_INST_CONST_SRC_SHIFT);
	case SRF_ATTRIB:
		src = ((shader->inst[1] & NV20_VP_INST_INPUT_SRC_MASK)
				>> NV20_VP_INST_INPUT_SRC_SHIFT);
		switch (src) {
		case NV20_VP_INST_INPUT_SRC_POS   : return SR_POSITION;
		case NV20_VP_INST_INPUT_SRC_COL0  : return SR_COL0;
		case NV20_VP_INST_INPUT_SRC_COL1  : return SR_COL1;
		case NV20_VP_INST_INPUT_SRC_TC(0) : return SR_TEXCOORD0;
		case NV20_VP_INST_INPUT_SRC_TC(1) : return SR_TEXCOORD1;
		case NV20_VP_INST_INPUT_SRC_TC(2) : return SR_TEXCOORD2;
		case NV20_VP_INST_INPUT_SRC_TC(3) : return SR_TEXCOORD3;
		default:
			return SR_UNKNOWN;
		}
	default:
		return -1;
	}
}

static int
NV20VPGetSourceNegate(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	
	src = shader->GetSourceHW(shader, merged, pos);

	return ((src & NV20_VP_SRC_REG_NEGATE) ? 1 : 0);
}

static int
NV20VPGetSourceAbs(ShaderFunc *shader, int merged, int pos)
{
	/* NV20 can't do ABS on sources?  Appears to be emulated with
	 * 	MAX reg, reg, -reg
	 */
	return 0;
}

static ShaderRegSwz
NV20VPGetSourceSwizzle(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	int swz;

	src = shader->GetSourceHW(shader, merged, pos);
	swz = (src & NV20_VP_SRC_REG_SWZ_ALL_MASK) >> NV20_VP_SRC_REG_SWZ_ALL_SHIFT;
	return NV20VPTXSwizzle(swz);
}

static int
NV20VPGetSourceIndexed(ShaderFunc *shader, int merged, int pos)
{
	/* I don't think NV20 can index into attribs, at least no GL
	 * extension is exposed that will allow it.
	 */
	if (shader->GetSourceFile(shader, merged, pos) != SRF_CONST)
		return 0;
	if (shader->inst[3] & NV20_VP_INST_INDEX_CONST)
		return 1;
	return 0;
}

static int
NV20VPGetAddressRegID(ShaderFunc *shader)
{
	/* Only 1 address reg */
	return 0;
}

static ShaderSwzComp
NV20VPGetAddressRegSwizzle(ShaderFunc *shader)
{
	/* Only A0.x available */
	return SSC_X;
}

void
NV20VPInitShaderFuncs(ShaderFunc *shader)
{
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_NOP, SOP_NOP, -1, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_MOV, SOP_MOV,  0, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_MUL, SOP_MUL,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_ADD, SOP_ADD,  0,  2, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_MAD, SOP_MAD,  0,  1,  2);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_DP3, SOP_DP3,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_DPH, SOP_DPH,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_DP4, SOP_DP4,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_DST, SOP_DST,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_MIN, SOP_MIN,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_MAX, SOP_MAX,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_SLT, SOP_SLT,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_SGE, SOP_SGE,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV20_VP_INST_OPCODE_ARL, SOP_ARL,  0,  1, -1);

	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_NOP, SOP_NOP, -1, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_RCP, SOP_RCP,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_RCC, SOP_RCC,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_RSQ, SOP_RSQ,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_EXP, SOP_EXP,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_LOG, SOP_LOG,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV20_VP_INST_OPCODE_LIT, SOP_LIT,  2, -1, -1);

	shader->GetOPTXRec				= NV20VPGetOPTXRec;
	shader->HasMergedInst			= NV20VPHasMergedInst;
	shader->IsLastInst				= NV20VPIsLastInst;
	shader->GetOffsetNext			= NV20VPGetOffsetNext;
	shader->GetOpcodeSlot			= NV20VPGetOpcodeSlot;
	shader->GetOpcode				= NV20VPGetOpcode;
	shader->GetOpcodeHW				= NV20VPGetOpcodeHW;
	shader->GetDestFile				= NV20VPGetDestFile;
	shader->GetDestID				= NV20VPGetDestID;
	shader->GetDestMask				= NV20VPGetDestMask;
	shader->GetSourceHW				= NV20VPGetSourceHW;
	shader->GetSourceFile			= NV20VPGetSourceFile;
	shader->GetSourceID				= NV20VPGetSourceID;
	shader->GetSourceNegate			= NV20VPGetSourceNegate;
	shader->GetSourceAbs			= NV20VPGetSourceAbs;
	shader->GetSourceSwizzle		= NV20VPGetSourceSwizzle;
	shader->GetSourceIndexed		= NV20VPGetSourceIndexed;
	shader->GetRelAddressRegID		= NV20VPGetAddressRegID;
	shader->GetRelAddressSwizzle	= NV20VPGetAddressRegSwizzle;
}

