#ifndef MMTRACE
#include <stdlib.h>
#endif

#include "shader_common.h"
#include "nv30_shader.h"

extern ShaderSwzComp NV20VP_TX_SWIZZLE[4];
extern ShaderRegSwz  NV20VPTXSwizzle(int hwswz);

static unsigned int
NV30VPGetOpcodeHW(ShaderFunc *shader, int slot)
{
	int op;

	if (slot) {
		op = (shader->inst[1] & NV30_VP_INST_SCA_OPCODEL_MASK)
			>> NV30_VP_INST_SCA_OPCODEL_SHIFT;
		op |= ((shader->inst[0] & NV30_VP_INST_SCA_OPCODEH_MASK)
			>> NV30_VP_INST_SCA_OPCODEH_SHIFT) << 4;
	} else {
		op = (shader->inst[1] & NV30_VP_INST_VEC_OPCODE_MASK)
			>> NV30_VP_INST_VEC_OPCODE_SHIFT;
	}
	
	return op;
}

static ShaderRegFile
NV30VPGetDestFile(ShaderFunc *shader, int merged)
{
	switch (shader->GetOpcode(shader, merged)) {
	case SOP_ARL:
	case SOP_ARR:
	case SOP_ARA:
		return SRF_ADDRESS;
	default:
	/*FIXME: This probably isn't correct.. */
		if ((shader->inst[3] & NV30_VP_INST_VDEST_WRITEMASK_MASK) != 0)
			return SRF_RESULT;
		if ((shader->inst[3] & NV30_VP_INST_SDEST_WRITEMASK_MASK) != 0)
			return SRF_RESULT;
		return SRF_TEMP;
	}
}

static unsigned int
NV30VPGetDestID(ShaderFunc *shader, int merged)
{
	int id;

	switch (shader->GetDestFile(shader, merged)) {
	case SRF_RESULT:
		id = ((shader->inst[3] & NV30_VP_INST_DEST_ID_MASK)
				>> NV30_VP_INST_DEST_ID_SHIFT);
		switch (id) {
		case NV30_VP_INST_DEST_POS  : return SR_POSITION;
		case NV30_VP_INST_DEST_COL0 : return SR_COL0;
		case NV30_VP_INST_DEST_COL1 : return SR_COL1;
		case NV30_VP_INST_DEST_TC(0): return SR_TEXCOORD0;
		case NV30_VP_INST_DEST_TC(1): return SR_TEXCOORD1;
		case NV30_VP_INST_DEST_TC(2): return SR_TEXCOORD2;
		case NV30_VP_INST_DEST_TC(3): return SR_TEXCOORD3;
		case NV30_VP_INST_DEST_TC(4): return SR_TEXCOORD4;
		case NV30_VP_INST_DEST_TC(5): return SR_TEXCOORD5;
		case NV30_VP_INST_DEST_TC(6): return SR_TEXCOORD6;
		case NV30_VP_INST_DEST_TC(7): return SR_TEXCOORD7;
		default:
			return -1;
		}
	case SRF_ADDRESS:
	case SRF_TEMP:
		return (shader->inst[0] & NV30_VP_INST_DEST_TEMP_ID_MASK)
			>> NV30_VP_INST_DEST_TEMP_ID_SHIFT;
	default:
		return -1;
	}
}

static unsigned int
NV30VPGetDestMask(ShaderFunc *shader, int merged)
{
	int hwmask, mask = 0;

	if (shader->GetDestFile(shader, merged) == SRF_RESULT)
		if (shader->GetOpcodeSlot(shader, merged))
			hwmask = (shader->inst[3] & NV30_VP_INST_SDEST_WRITEMASK_MASK)
				>> NV30_VP_INST_SDEST_WRITEMASK_SHIFT;
		else
			hwmask = (shader->inst[3] & NV30_VP_INST_VDEST_WRITEMASK_MASK)
				>> NV30_VP_INST_VDEST_WRITEMASK_SHIFT;
	else
		if (shader->GetOpcodeSlot(shader, merged))
			hwmask = (shader->inst[3] & NV30_VP_INST_STEMP_WRITEMASK_MASK)
				>> NV30_VP_INST_STEMP_WRITEMASK_SHIFT;
		else
			hwmask = (shader->inst[3] & NV30_VP_INST_VTEMP_WRITEMASK_MASK)
				>> NV30_VP_INST_VTEMP_WRITEMASK_SHIFT;

	if (hwmask & (1<<3)) mask |= SMASK_X;
	if (hwmask & (1<<2)) mask |= SMASK_Y;
	if (hwmask & (1<<1)) mask |= SMASK_Z;
	if (hwmask & (1<<0)) mask |= SMASK_W;

	return mask;
}

int
NV30VPGetSourceID(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;

	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_TEMP:
		src = shader->GetSourceHW(shader, merged, pos);
		return ((src & NV30_VP_SRC_REG_TEMP_ID_MASK) >>
				NV30_VP_SRC_REG_TEMP_ID_SHIFT);
	case SRF_CONST:
		return ((shader->inst[1] & NV30_VP_INST_CONST_SRC_MASK)
				>> NV30_VP_INST_CONST_SRC_SHIFT);
	case SRF_ATTRIB:
		src = ((shader->inst[1] & NV30_VP_INST_INPUT_SRC_MASK)
				>> NV30_VP_INST_INPUT_SRC_SHIFT);
		switch (src) {
		case NV30_VP_INST_IN_POS   : return SR_POSITION;
		case NV30_VP_INST_IN_COL0  : return SR_COL0;
		case NV30_VP_INST_IN_COL1  : return SR_COL1;
		case NV30_VP_INST_IN_TC(0) : return SR_TEXCOORD0;
		case NV30_VP_INST_IN_TC(1) : return SR_TEXCOORD1;
		case NV30_VP_INST_IN_TC(2) : return SR_TEXCOORD2;
		case NV30_VP_INST_IN_TC(3) : return SR_TEXCOORD3;
		case NV30_VP_INST_IN_TC(4) : return SR_TEXCOORD4;
		case NV30_VP_INST_IN_TC(5) : return SR_TEXCOORD5;
		case NV30_VP_INST_IN_TC(6) : return SR_TEXCOORD6;
		case NV30_VP_INST_IN_TC(7) : return SR_TEXCOORD7;
		default:
			return SR_UNKNOWN;
		}
	default:
		return -1;
	}
}

static int
NV30VPGetSourceAbs(ShaderFunc *shader, int merged, int pos)
{
	struct _op_xlat *opr;
	static unsigned int abspos[3] = {
		NV30_VP_INST_SRC0_ABS,
		NV30_VP_INST_SRC1_ABS,
		NV30_VP_INST_SRC2_ABS,
	};

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1 || opr->srcpos[pos] > 2)
		return 0;

	return ((shader->inst[0] & abspos[opr->srcpos[pos]]) ? 1 : 0);
}

static int
NV30VPGetRelAddressRegID(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV30_VP_INST_ADDR_REG_SELECT_1) ? 1 : 0);
}

static ShaderSwzComp
NV30VPGetRelAddressSwizzle(ShaderFunc *shader)
{
	ShaderSwzComp swz;

	swz = NV20VP_TX_SWIZZLE[(shader->inst[0] & NV30_VP_INST_ADDR_SWZ_MASK)
		>> NV30_VP_INST_ADDR_SWZ_SHIFT];
	return swz;
}

static int
NV30VPSupportsConditional(ShaderFunc *shader)
{
	/*FIXME: Is this true of all ops? */
	return 1;
}

static int
NV30VPGetConditionUpdate(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV30_VP_INST_COND_UPDATE_ENABLE) ? 1 : 0);
}

static int
NV30VPGetConditionTest(ShaderFunc *shader)
{
	int op;

	/* The condition test is unconditionally enabled on some
	 * instructions. ie: the condition test bit does *NOT* have
	 * to be set.
	 *
	 * FIXME: check other relevant ops for this situation.
	 */
	op = shader->GetOpcodeHW(shader, 1);
	switch (op) {
	case NV30_VP_INST_OP_BRA:
		return 1;
	default:
		return ((shader->inst[0] & NV30_VP_INST_COND_TEST_ENABLE) ? 1 : 0);
	}
}

static ShaderCond
NV30VPGetCondition(ShaderFunc *shader)
{
	int cond;

	cond = ((shader->inst[0] & NV30_VP_INST_COND_MASK)
			>> NV30_VP_INST_COND_SHIFT);

	switch (cond) {
	case NV30_VP_INST_COND_FL: return SC_FL;
	case NV30_VP_INST_COND_LT: return SC_LT;
	case NV30_VP_INST_COND_EQ: return SC_EQ;
	case NV30_VP_INST_COND_LE: return SC_LE;
	case NV30_VP_INST_COND_GT: return SC_GT;
	case NV30_VP_INST_COND_NE: return SC_NE;
	case NV30_VP_INST_COND_GE: return SC_GE;
	case NV30_VP_INST_COND_TR: return SC_TR;
	default:
		return SC_UNKNOWN;
	}
}

static ShaderRegSwz
NV30VPGetCondRegSwizzle(ShaderFunc *shader)
{
	int swz;

	swz = (shader->inst[0] & NV30_VP_INST_COND_SWZ_ALL_MASK)
		>> NV30_VP_INST_COND_SWZ_ALL_SHIFT;
	return NV20VPTXSwizzle(swz);
}

static int
NV30VPGetCondRegID(ShaderFunc *shader)
{
	return 0;
}


static int
NV30VPGetBranch(ShaderFunc *shader)
{
	return ((shader->inst[2] & NV30_VP_INST_IADDR_MASK)
			>> NV30_VP_INST_IADDR_SHIFT);
}

void
NV30VPInitShaderFuncs(ShaderFunc *shader)
{
	/* Inherit NV20 code, a lot of it is the same */
	NV20VPInitShaderFuncs(shader);

	/* Increase max valid opcode ID, and add new instructions */
	NVVP_TX_VOP_COUNT = NVVP_TX_SOP_COUNT = 32;

	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_FRC, SOP_FRC,  0, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_FLR, SOP_FLR,  0, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SEQ, SOP_SEQ,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SFL, SOP_SFL,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SGT, SOP_SGT,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SLE, SOP_SLE,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SNE, SOP_SNE,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_STR, SOP_STR,  0,  1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_SSG, SOP_SSG,  0, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_ARR, SOP_ARR,  0, -1, -1);
	MOD_OPCODE(NVVP_TX_VOP, NV30_VP_INST_OP_ARA, SOP_ARA,  3, -1, -1);

	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_BRA, SOP_BRA, -1, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_CAL, SOP_CAL, -1, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_RET, SOP_RET, -1, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_LG2, SOP_LG2,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_EX2, SOP_EX2,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_SIN, SOP_SIN,  2, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV30_VP_INST_OP_COS, SOP_COS,  2, -1, -1);

	shader->GetOpcodeHW				= NV30VPGetOpcodeHW;

	shader->GetDestFile				= NV30VPGetDestFile;
	shader->GetDestID				= NV30VPGetDestID;
	shader->GetDestMask				= NV30VPGetDestMask;

	shader->GetSourceID	= NV30VPGetSourceID;
	shader->GetSourceAbs			= NV30VPGetSourceAbs;

	shader->GetRelAddressRegID		= NV30VPGetRelAddressRegID;
	shader->GetRelAddressSwizzle	= NV30VPGetRelAddressSwizzle;

	shader->SupportsConditional		= NV30VPSupportsConditional;
	shader->GetConditionUpdate		= NV30VPGetConditionUpdate;
	shader->GetConditionTest		= NV30VPGetConditionTest;
	shader->GetCondition			= NV30VPGetCondition;
	shader->GetCondRegSwizzle		= NV30VPGetCondRegSwizzle;
	shader->GetCondRegID			= NV30VPGetCondRegID;

	shader->GetBranch				= NV30VPGetBranch;
}

